import dataclasses


@dataclasses.dataclass
class Position:
    lineno: int
    start_col: int
    end_col: int

    @property
    def col_range(self) -> int:
        return self.end_col - self.start_col


@dataclasses.dataclass
class ErrorLine:
    filename: str
    line: str
    pos: Position
    message: str

    def indicators(self) -> str:
        offset1 = self.pos.start_col + 1
        offset2 = self.pos.col_range - 4
        return f"{offset1 * ' '}^{offset2 * '-'}^"


def collate_errors(errors: list[ErrorLine]) -> dict[str, dict[int, list[ErrorLine]]]:
    """
    Organize errors first by filename, secondarily by line number.
    """
    by_filename: dict[str, list[ErrorLine]] = {}
    for e in errors:
        by_filename.setdefault(e.filename, [])
        by_filename[e.filename].append(e)

    collated: dict[str, dict[int, list[ErrorLine]]] = {}
    for filename, file_errors in by_filename.items():
        by_lineno: dict[int, list[ErrorLine]] = {}
        for e in file_errors:
            by_lineno.setdefault(e.pos.lineno, [])
            by_lineno[e.pos.lineno].append(e)
        collated[filename] = by_lineno

    return collated



import re
import typing as t

from ._model import ErrorLine, Position

BAD_BACKTICK_PAT = re.compile(r"(^|\s)`[^`]+`([.,?!\s]|$)", flags=re.MULTILINE)
CODE_BLOCK_START_PAT = re.compile(r"^\s*\.\.\s+code-block::")


class BacktickChecker:
    def __init__(self, filename: str, contents: list[str]) -> None:
        self.filename = filename
        self.contents = contents

    def iter_errors(self) -> t.Iterator[ErrorLine]:
        code_block_detector = _CodeBlockDetector()

        for lineno, line in enumerate(self.contents, 1):
            # consume/skip lines while in a code-block
            if code_block_detector.in_code_block and not code_block_detector.block_end(
                line
            ):
                continue

            # if we are on the first line of a code-block, start skipping!
            if code_block_detector.set_start(line):
                continue

            line = line.rstrip("\n")
            yield from self._iter_line_errors(line, lineno)

    def _iter_line_errors(self, line: str, lineno: int) -> t.Iterator[ErrorLine]:
        for m in BAD_BACKTICK_PAT.finditer(line):
            pos = Position(lineno=lineno, start_col=m.start(), end_col=m.end())
            yield ErrorLine(
                filename=self.filename,
                line=line,
                pos=pos,
                message="found what appear to be markdown backticks",
            )


class _CodeBlockDetector:
    def __init__(self) -> None:
        self.in_code_block = False
        self._indent = 0

    def set_start(self, line: str) -> bool:
        """
        Try to start a code-block.

        If the line begins a new code-block, return true. Otherwise, false.
        """
        if CODE_BLOCK_START_PAT.match(line):
            self.in_code_block = True
            self._indent = _get_indent(line)
            return True
        return False

    def block_end(self, line: str) -> bool:
        """
        Try to end a code-block.

        If the line ends the current code-block, return true. Otherwise, false.
        """
        # skip empty lines
        if not line.strip():
            return False
        if _get_indent(line) <= self._indent:
            self.in_code_block = False
            return True
        return False


def _get_indent(line: str) -> int:
    return len(line) - len(line.lstrip())



import sys
import typing as t

import click

from ._checkers import BacktickChecker
from ._model import ErrorLine, collate_errors


@click.group
def main():
    """a linter for your rst mistakes"""


@main.command
@click.argument("files", required=True, nargs=-1, type=click.File("r"))
def bad_backticks(files: t.IO[bytes]):
    errors: list[ErrorLine] = []
    for file in files:
        errors.extend(BacktickChecker(file.name, file.readlines()).iter_errors())

    if errors:
        collated = collate_errors(errors)
        for filename, by_lineno in collated.items():
            print(f"encountered errors in {filename}:")
            for lineno, line_errors in by_lineno.items():
                prefix = f"  on line {lineno}: "
                prefixlen = len(prefix)
                line = line_errors[0].line
                print(f"{prefix}{line}")
                for e in line_errors:
                    indicators = e.indicators()
                    message = (" " * (len(e.line) - len(indicators) + 2)) + e.message
                    print(f"{' ' * prefixlen}{indicators}{message}")
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()
